"""
Functions related to masking Websocket frames.
https://tools.ietf.org/html/rfc6455#section-5.3

"""

import os
from functools import partial

import six


make_masking_key = partial(os.urandom, 4)


if six.PY2:
    _XOR_TABLE = [b''.join(chr(a ^ b) for a in range(256)) for b in range(256)]
else:
    _XOR_TABLE = [bytes(a ^ b for a in range(256)) for b in range(256)]


def mask_payload(masking_key, data):
    """XOR mask bytes.

    `masking_key` should be bytes.
    `data` should be a bytearray, and is mutated.

    """
    a, b, c, d = (_XOR_TABLE[n] for n in bytearray(masking_key))
    data[::4] = data[::4].translate(a)
    data[1::4] = data[1::4].translate(b)
    data[2::4] = data[2::4].translate(c)
    data[3::4] = data[3::4].translate(d)



"""
Coroutine Parser
================

This little but of 'magic' makes parsing a stream of bytes from a socket
a relatively pain-free exercise.

"""

from __future__ import unicode_literals


class ParseError(Exception):
    """Stream failed to parse."""


class ParseEOF(ParseError):
    """End of Stream."""


class ParseOverflow(Exception):
    """Extra bytes in feed after parser completed."""


class _Awaitable(object):
    """An operation that effectively suspends the coroutine."""

    # Analogous to Python3 asyncio concept
    __slots__ = []

    def validate(self, chunk):
        """Raise any ParseErrors"""


class _ReadBytes(_Awaitable):
    """Reads a fixed number of bytes."""

    __slots__ = ["remaining"]

    def __init__(self, count):
        self.remaining = count


class _ReadUtf8(_ReadBytes):
    """Reads a fixed number of bytes, validates utf-8."""

    __slots__ = ["utf8_validator"]

    def __init__(self, count, utf8_validator):
        self.remaining = count
        self.utf8_validator = utf8_validator

    def validate(self, data):
        valid, _, _, _ = self.utf8_validator.validate(bytes(data))
        if not valid:
            raise ParseError("invalid utf8")


class _ReadUntil(_Awaitable):
    """Read until a separator."""

    __slots__ = ["sep", "max_bytes"]

    def __init__(self, sep, max_bytes=None):
        self.sep = sep
        self.max_bytes = max_bytes

    def check_length(self, pos):
        """Check the length is within max bytes."""
        if self.max_bytes is not None and pos > self.max_bytes:
            raise ParseError("expected {!r}".format(self.sep))


class Parser(object):
    """
    Coroutine based stream parser.

    Splits a steam of arbitrary sequences of bytes in to logical
    objects. The `feed` method will yield any results of parsing.

    Here's an example of use::

        while True:
            data = sock.recv(1024)
            for obj in parser.feed(data):
                self.on_obj(obj)

    The coroutine magic allows parsers to be non-blocking while still
    implemented in a simple procedural manner.

    """

    def __init__(self):
        self._gen = None
        self._awaiting = None
        self._buffer = bytearray()  # Buffer for reads
        self._eof = False
        self._exhausted = False
        self.reset()

    read = _ReadBytes
    read_utf8 = _ReadUtf8
    read_until = _ReadUntil

    def __del__(self):
        self.close()

    @property
    def is_eof(self):
        return self._eof

    def reset(self):
        """Reset the parser, so it may be used on a fresh stream."""
        self._gen = self.parse()
        self._awaiting = next(self._gen)

    def close(self):
        """Close the parser."""
        if self._gen is not None:
            self._gen.close()
            self._gen = None

    def feed(self, data):
        """
        Called with data (bytes), will yield 0 or more objects parsed
        from the stream.

        :param bytes data: Data to parse.

        """

        def _check_length(pos):
            try:
                self._awaiting.check_length(pos)
            except ParseError as error:
                self._awaiting = self._gen.throw(error)

        if self._exhausted:
            raise ParseOverflow("extra bytes in feed(); {!r}".format(data[:100]))
        if self._eof:
            raise ParseEOF(
                "end of file reached; feed() has previously been called with empty bytes"
            )
        if not data:
            self._eof = True
            self._gen.throw(ParseEOF("unexpected eof of file"))

        try:
            _buffer = self._buffer
            pos = 0
            while pos < len(data):
                # Awaiting a read of a fixed number of bytes
                if isinstance(self._awaiting, _ReadBytes):
                    # This many bytes left to read
                    remaining = self._awaiting.remaining
                    # Bite off remaining bytes
                    chunk = data[pos : pos + remaining]
                    chunk_size = len(chunk)
                    pos += chunk_size
                    try:
                        # Validate new data
                        self._awaiting.validate(chunk)
                    except ParseError as error:
                        # Raises an exception in parse()
                        self._awaiting = self._gen.throw(error)
                    # Add to buffer
                    _buffer.extend(chunk)
                    remaining -= chunk_size
                    if remaining:
                        # Await more bytes
                        self._awaiting.remaining = remaining
                    else:
                        # Send to coroutine, get new 'awaitable'
                        self._awaiting = self._gen.send(_buffer[:])
                        del _buffer[:]

                # Awaiting a read until a terminator
                elif isinstance(self._awaiting, _ReadUntil):
                    # Reading to separator
                    chunk = data[pos:]
                    _buffer.extend(chunk)
                    sep = self._awaiting.sep
                    sep_index = _buffer.find(sep)

                    if sep_index == -1:
                        # Separator not found, advance position
                        pos += len(chunk)
                        _check_length(len(_buffer))
                    else:
                        # Found separator
                        # Get data prior to and including separator
                        sep_index += len(sep)
                        _check_length(sep_index)
                        # Reset data, to continue parsing
                        data = _buffer[sep_index:]
                        pos = 0
                        # Send bytes to coroutine, get new 'awaitable'
                        self._awaiting = self._gen.send(_buffer[:sep_index])
                        del _buffer[:]

                # Yield any non-awaitables...
                while not isinstance(self._awaiting, _Awaitable):
                    yield self._awaiting
                    self._awaiting = next(self._gen)
        except StopIteration:
            self._exhausted = True
            if pos < len(data):
                raise ParseOverflow(
                    "extra bytes in feed(); {!r}".format(data[pos:][:100])
                )

    def parse(self):
        """
        A generator to parse incoming stream.

        Yield the result of `self.read` to read n bytes from the stream.
        Yield any parsed objects

        Here's an example::

            size = struct.unpack('<I', yield self.read(4))
            data = yield self.read(size)
            yield data.decode()

        """
        yield


if __name__ == "__main__":  # pragma: no cover

    class TestParser(Parser):
        def parse(self):
            data = yield self.read_until(b"\r\n\r\n")
            yield data
            data = yield self.read(1)
            yield data
            data = yield self.read(2)
            yield data
            data = yield self.read(4)
            yield data
            data = yield self.read(2)
            yield data

    parser = TestParser()
    for b in (
        b"head",
        b"ers: example",
        b"\r\n",
        b"\r\n",
        b"12",
        b"34",
        b"5",
        b"678",
        b"9",
        b"9",
        b"9",
    ):
        for frame in parser.feed(b):
            print(repr(frame))



from __future__ import unicode_literals


def parse_extension(extension):
    """Parse a single extension in to an extension token and a dict
    of options.

    """
    # Naive http header parser, works for current extensions
    tokens = [token.strip() for token in extension.split(';')]
    extension_token = tokens[0]
    options = {}
    for token in tokens[1:]:
        key, sep, value = token.partition('=')
        value = value.strip().strip('"')
        options[key.strip()] = value
    return extension_token, options



from __future__ import unicode_literals

import json
import time


class Event(object):
    """Base class for a websocket 'event'."""
    __slots__ = ['received_time']

    def __init__(self):
        self.received_time = time.time()

    def __repr__(self):
        return "{}()".format(self.__class__.__name__)

    @classmethod
    def _summarize_bytes(cls, data, max_len=24):
        """Avoid spamming logs by truncating byte strings in repr."""
        if len(data) > max_len:
            return "{!r} + {} bytes".format(
                data[:max_len],
                len(data) - max_len
            )
        return repr(data)

    @classmethod
    def _summarize_text(cls, text, max_len=24):
        """Avoid spamming logs by truncating text."""
        if len(text) > max_len:
            return "{!r} + {} chars".format(
                text[:max_len],
                len(text) - max_len
            )
        return repr(text)


class Poll(Event):
    """A generated poll event."""
    name = 'poll'


class Connecting(Event):
    """
    Generated prior to establishing a websocket connection to a server.

    :param url: The websocket URL the websocket is connecting to.

    """
    __slots__ = ['url']
    name = 'connecting'

    def __init__(self, url):
        self.url = url
        super(Connecting, self).__init__()

    def __repr__(self):
        return "{}(url='{}')".format(self.__class__.__name__, self.url)


class ConnectFail(Event):
    """
    Generate when Lomond was unable to connect to a Websocket server.

    :param reason: A short description of the reason for the
        failure.
    :type reason: str

    """
    __slots__ = ['reason']
    name = 'connect_fail'

    def __init__(self, reason):
        self.reason = reason
        super(ConnectFail, self).__init__()

    def __repr__(self):
        return "{}(reason='{}')".format(
            self.__class__.__name__,
            self.reason,
        )


class Connected(Event):
    """Generated when Lomond has connected to a server but not yet
    negotiated the websocket upgrade.

    :param str url: The websocket URL connected to.
    :param str proxy: The proxy URL connected to (or None).

    """
    __slots__ = ['url', 'proxy']
    name = 'connected'

    def __init__(self, url, proxy=None):
        self.url = url
        self.proxy = proxy
        super(Connected, self).__init__()

    def __repr__(self):
        _class = self.__class__.__name__
        return (
            "{}(url='{}')".format(_class, self.url)
            if self.proxy is None else
            "{}(url='{}', proxy='{}')".format(
                _class, self.url, self.proxy
            )
        )


class Rejected(Event):
    """Server rejected WS connection."""
    __slots__ = ['response', 'reason']
    name = 'rejected'

    def __init__(self, response, reason):
        """
        Generated when Lomond is connected to the server, but the
        websocket upgrade failed.

        :param response: The response returned by the server.
        :param str reason: A description of why the connection was
            rejects.

        """
        self.response = response
        self.reason = reason
        super(Rejected, self).__init__()

    def __repr__(self):
        return "{}(response={!r}, reason='{}')".format(
            self.__class__.__name__,
            self.response,
            self.reason
        )


class Ready(Event):
    """Generated when Lomond has connected to the server,
    and successfully negotiated the websocket upgrade.

    :param response: A :class:`~lomond.response.Response` object.
    :param str protocol: A websocket protocol or ``None`` if no protocol
        was supplied.
    :param set extensions: A set of negotiated websocket extensions.
        Currently only the ``'permessage-deflate'`` extension is supported.

    """
    __slots__ = ['response', 'protocol', 'extensions']
    name = 'ready'

    def __init__(self, response, protocol, extensions):
        self.response = response
        self.protocol = protocol
        self.extensions = extensions
        super(Ready, self).__init__()

    def __repr__(self):
        return '{}(response={!r}, protocol={!r}, extensions={!r})'.format(
            self.__class__.__name__,
            self.response,
            self.protocol,
            self.extensions
        )


class ProtocolError(Event):
    """Generated when the server deviates from the protocol.

    :param str error: A description of the error.
    :param bool critical: Indicates if the error is considered
        'critical'. If ``True``, Lomond will disconnect immediately.
        If ``False``, Lomond will send a close message to the server.

    """
    __slots__ = ['error', 'critical']
    name = 'protocol_error'

    def __init__(self, error, critical):
        self.error = error
        self.critical = critical
        super(ProtocolError, self).__init__()

    def __repr__(self):
        return "{}(error='{}', critical={!r})".format(
            self.__class__.__name__,
            self.error,
            self.critical
        )


class Unresponsive(Event):
    """The server has not responding to pings within `ping_timeout`
    seconds.

    Will be followed by a :class:`~lomond.events.Disconnected` event.

    """
    name = 'unresponsive'


class Disconnected(Event):
    """Generated when a websocket connection has
    been dropped.

    :param str reason: A description of why the websocket was closed.
    :param bool graceful: Flag indicating if the connection was dropped
        gracefully (`True`), or disconnected due to a socket failure
        (`False`) or other problem.

    """
    __slots__ = ['graceful', 'reason']
    name = 'disconnected'

    def __init__(self, reason='closed', graceful=False):
        self.reason = reason
        self.graceful = graceful
        super(Disconnected, self).__init__()

    def __repr__(self):
        return "{}(reason='{}', graceful={!r})".format(
            self.__class__.__name__,
            self.reason,
            self.graceful
        )


class Closed(Event):
    """Generated when the websocket was closed. The websocket may no
    longer send packets after this event has been received. This event
    will be followed by :class:`~lomond.events.Disconnected`.

    :param code: The closed code returned from the server.
    :param str reason: An optional description why the websocket was
        closed, as returned from the server.

    """
    __slots__ = ['code', 'reason']
    name = 'closed'

    def __init__(self, code, reason):
        self.code = code
        self.reason = reason
        super(Closed, self).__init__()

    def __repr__(self):
        return '{}(code={!r}, reason={!r})'.format(
            self.__class__.__name__,
            self.code,
            self.reason,
        )


class Closing(Event):
    """Generated when the server is closing the connection.

    No more messages will be received from the server, but you may still
    send messages while handling this event. A
    :class:`~lomond.events.Disconnected` event should be generated
    shortly after this event.

    :param code: The closed code returned from the server.
    :param str reason: An optional description why the websocket was
        closed, as returned from the server.

    """
    __slots__ = ['code', 'reason']
    name = 'closing'

    def __init__(self, code, reason):
        self.code = code
        self.reason = reason
        super(Closing, self).__init__()

    def __repr__(self):
        return '{}(code={!r}, reason={!r})'.format(
            self.__class__.__name__,
            self.code,
            self.reason,
        )


class UnknownMessage(Event):
    """
    An application message was received, with an unknown opcode.

    """
    __slots__ = ['message']
    name = 'unknown'

    def __init__(self, message):
        self.message = message
        super(UnknownMessage, self).__init__()


class Ping(Event):
    """Generated when Lomond received a ping packet from the server.

    :param bytes data: Ping payload data.

    """
    __slots__ = ['data']
    name = 'ping'

    def __init__(self, data):
        self.data = data
        super(Ping, self).__init__()

    def __repr__(self):
        return "{}(data={!r})".format(self.__class__.__name__, self.data)


class Pong(Event):
    """Generated when Lomond receives a pong packet from the server.

    :param bytes data: The pong payload data.

    """
    __slots__ = ['data']
    name = 'pong'

    def __init__(self, data):
        self.data = data
        super(Pong, self).__init__()

    def __repr__(self):
        return "{}(data={!r})".format(self.__class__.__name__, self.data)


class Text(Event):
    """Generated when Lomond receives a text message from the server.

    :param str text: The text payload.

    """
    __slots__ = ['text', '_json']
    name = 'text'

    def __init__(self, text):
        self.text = text
        self._json = None
        super(Text, self).__init__()

    @property
    def json(self):
        """Text decoded as JSON.

        Calls ``json.loads`` to decode the ``text`` attribute, and may
        throw the same exceptions if the text is not valid json.

        """
        if self._json is None:
            self._json = json.loads(self.text)
        return self._json

    def __repr__(self):
        return "{}(text={})".format(
            self.__class__.__name__,
            self._summarize_text(self.text)
        )


class Binary(Event):
    """Generated when Lomond receives a binary message from the server.

    :param bytes data: The binary payload.

    """
    __slots__ = ['data']
    name = 'binary'

    def __init__(self, data):
        self.data = data
        super(Binary, self).__init__()

    def __repr__(self):
        return "{}(data={})".format(
            self.__class__.__name__,
            self._summarize_bytes(self.data)
        )


class BackOff(Event):
    """Generated when a persistent connection has to wait before re-
    attempting a connection.

    :param float delay: The delay (in seconds) before Lomond will re-
        attempt to connect.

    """
    __slots__ = ['delay']
    name = 'back_off'

    def __init__(self, delay):
        self.delay = delay
        super(BackOff, self).__init__()

    def __repr__(self):
        return "{}(delay={:0.1f})".format(
            self.__class__.__name__,
            self.delay
        )



from __future__ import unicode_literals

import base64

from .parser import Parser, ParseError
from .response import Response


class ProxyFail(Exception):
    """An error with the proxy."""
    # An internal exception


class ProxyResponse(Response):
    """The response received from the proxy."""


def build_request(host, port, proxy_username=None, proxy_password=None):
    """Build a request to the proxy."""
    request = [
        'CONNECT {}:{} HTTP/1.1'.format(host, port).encode('utf-8')
    ]
    headers = [
        (b'Host', host.encode('utf-8')),
        (b'Proxy-Connection', b'keep-alive'),
        (b'Connection', b'keep-alive'),
    ]

    if proxy_username:
        credentials = (
            proxy_username
            if proxy_password is None else
            '{}:{}'.format(proxy_username, proxy_password)
        ).encode('utf-8')
        b64_credentials = base64.standard_b64encode(credentials)
        headers.append(
            (b'Proxy-Authorization:', b'Basic ' + b64_credentials)
        )

    for header, value in headers:
        request.append(header + b': ' + value)

    request.append(b'\r\n')
    request_bytes = b'\r\n'.join(request)
    return request_bytes


class ProxyParser(Parser):
    """Parser for communication with a SOCKS proxy."""

    def parse(self):
        try:
            headers_data = yield self.read_until(b'\r\n\r\n', max_bytes=16 * 1024)
        except ParseError as error:
            raise ProxyFail('proxy parse fail; {}', error)
        response = ProxyResponse(headers_data)
        if response.status_code != 200:
            raise ProxyFail(
                'proxy fail; {} {}',
                response.status_code,
                response.status
            )
        yield response



from lomond import WebSocket

websocket = WebSocket('wss://ws-feed.gdax.com')

for event in websocket:
    if event.name == "ready":
        websocket.send_json(
            type='subscribe',
            product_ids=['BTC-USD'],
            channels=['ticker']
        )
    elif event.name == "text":
        print(event.json)



from __future__ import unicode_literals

import zlib

from six import PY2

from .errors import CompressionParameterError


class Deflate(object):
    """Compress with the Deflate algorithm."""

    def __init__(
        self, decompress_wbits, compress_wbits, reset_decompress, reset_compress
    ):
        self.decompress_wbits = decompress_wbits
        self.compress_wbits = compress_wbits
        self.reset_decompress = reset_decompress
        self.reset_compress = reset_compress
        self.reset_decompressor()
        self.reset_compressor()

    def __repr__(self):
        return "Deflate({}, {}, {}, {})".format(
            "decompress_wbits={!r}".format(self.decompress_wbits),
            "compress_wbits={!r}".format(self.compress_wbits),
            "reset_decompress={!r}".format(self.reset_decompress),
            "reset_compress={!r}".format(self.reset_compress),
        )

    def reset_compressor(self):
        """Reset the compressor for the next frame."""
        self._compressobj = zlib.compressobj(
            zlib.Z_DEFAULT_COMPRESSION,
            zlib.DEFLATED,
            -max(9, self.compress_wbits)
        )

    def reset_decompressor(self):
        """Reset the decompressor for the next frame."""
        self._decompressobj = zlib.decompressobj(-self.decompress_wbits)

    @classmethod
    def from_options(cls, options):
        """Build object from options dict."""
        decompress_wbits = cls.get_wbits(options, "server_max_window_bits")
        compress_wbits = cls.get_wbits(options, "client_max_window_bits")
        reset_decompress = "server_no_context_takeover" in options
        reset_compress = "client_no_context_takeover" in options
        deflate = Deflate(
            decompress_wbits, compress_wbits, reset_decompress, reset_compress
        )
        return deflate

    @classmethod
    def get_wbits(cls, options, key):
        """Parse wbits from options."""
        _wbits = options.get(key, "15")
        try:
            wbits = int(_wbits)
        except ValueError:
            raise CompressionParameterError("{} is not an integer".format(key))
        if wbits < 8 or wbits > 15:
            raise CompressionParameterError(
                "{}={} is invalid".format(key, wbits)
            )
        return wbits

    def decompress(self, frames):
        """Decompress payload, returned decompressed data."""
        if PY2:
            data = [
                self._decompressobj.decompress(bytes(frame.payload))
                for frame in frames
            ]
        else:
            data = [
                self._decompressobj.decompress(frame.payload)
                for frame in frames
            ]

        data.append(self._decompressobj.decompress(b"\x00\x00\xff\xff"))
        payload = b''.join(data)
        if self.reset_decompress:
            self.reset_decompressor()
        return payload

    def compress(self, payload):
        """Compress payload, return compressed data."""
        if PY2:
            payload = bytes(payload)
        data = (
            self._compressobj.compress(payload)
            + self._compressobj.flush(zlib.Z_SYNC_FLUSH)
        )[:-4]
        if self.reset_compress:
            self.reset_compressor()
        return data



"""
The session manages the mechanics of receiving and sending data over
the websocket.

"""

from __future__ import division
from __future__ import print_function
from __future__ import unicode_literals

import logging
import math
import socket
import ssl
import threading
import time

from six.moves.urllib.parse import urlparse

from .frame import Frame
from . import errors
from . import events
from . import proxy
from . import selectors


HAS_SNI = hasattr(ssl, 'SSLContext') and getattr(ssl, 'HAS_SNI', False)
log = logging.getLogger('lomond')


class _SocketFail(Exception):
    """Used internally to respond to socket fails."""


class _ForceDisconnect(Exception):
    """Used internally when the close timeout is tripped."""


class WebsocketSession(object):
    """Manages the mechanics of running the websocket."""
    _selector_cls = selectors.PlatformSelector

    BUFFER_SIZE = 64 * 1024

    def __init__(self, websocket):
        self.websocket = websocket
        self._address = (websocket.host, websocket.port)
        self._lock = threading.Lock()
        self._sock = None
        self._poll_start = None
        self._next_ping = None
        self._last_pong = None
        self._start_time = None
        self._ready = False
        self._buffer = bytearray(self.BUFFER_SIZE)

    def __repr__(self):
        return "<ws-session '{}'>".format(self.websocket.url)

    @property
    def session_time(self):
        """Get the time (in seconds) since the WebSocket session started."""
        return (
            0.0
            if self._start_time is None else
            time.time() - self._start_time
        )

    def close(self):
        """Close the websocket, if it is open."""
        self._close_socket()
        self._sock = None

    def force_disconnect(self):
        """Force the socket to disconnect."""
        raise _ForceDisconnect()

    def write(self, data):
        """Send raw data."""
        with self._lock:
            if self._sock is None:
                log.debug('WebSocket unavailable; data not sent')
                raise errors.WebSocketUnavailable('not connected')
            if self.websocket.is_closed:
                log.debug('WebSocket closed; data not sent')
                raise errors.WebSocketClosed('data not sent')
            if self.websocket.is_closing:
                log.debug('WebSocket closing; data not sent')
                raise errors.WebSocketClosing('data not sent')
            try:
                self._sock.sendall(data)
            except socket.error as error:
                log.debug('WebSocket send error; %s', error)
                raise errors.TransportFail(
                    'socket fail; {}', error
                )
            except Exception as error:
                log.warning('WebSocket send error; %s', error)
                raise errors.TransportFail(
                    'socket error; {}', error
                )

    def send(self, opcode, data):
        """Send a WS Frame."""
        frame = Frame(opcode, payload=bytearray(data))
        self.write(frame.to_bytes())
        log.debug(' SRV <- CLI : %r', frame)

    def send_compressed(self, opcode, data):
        """Send a compressed WS Frame."""
        frame = Frame(opcode, payload=bytearray(data), rsv1=1)
        self.write(frame.to_bytes())
        log.debug(' SRV <- CLI : %r', frame)

    @classmethod
    def _socket_fail(cls, msg, *args, **kwargs):
        """Raises a socket fail error to exit select loop."""
        _msg = msg.format(*args, **kwargs)
        log.debug(_msg)
        raise _SocketFail(_msg)

    def _connect_sock(self, host, port, ssl=False):
        sock = None
        try:
            addr_info = socket.getaddrinfo(
                host, port, socket.AF_UNSPEC, socket.SOCK_STREAM
            )
        except socket.error as error:
            self._socket_fail('unable to connect; {}', error)
        for res in addr_info:
            af, socktype, proto, canonname, sa = res
            try:
                sock = socket.socket(af, socktype, proto)
            except socket.error as error:
                log.debug('unable to create socket; %s', error)
                sock = None
                continue
            sock.setsockopt(socket.IPPROTO_TCP, socket.TCP_NODELAY, 1)
            sock.settimeout(30)  # TODO: make a parameter for this?
            if ssl:
                log.debug('wrapping socket')
                sock = self._wrap_socket(sock, host)
            try:
                sock.connect(sa)
            except socket.error as error:
                log.debug('socket error connecting to %r; %s', sa, error)
                sock.close()
                sock = None
                continue
            break
        if sock is None:
            self._socket_fail('unable to connect')
        return sock

    def _connect_proxy(self, proxy_url):
        """Connect to a http proxy, return socket."""
        _proxy_url = urlparse(proxy_url)
        _port = (
            int(_proxy_url.port)
            if _proxy_url.port else
            (443 if _proxy_url.scheme == 'https' else 80)
        )
        try:
            sock = self._connect_sock(
                _proxy_url.hostname, _port,
                ssl=_proxy_url.scheme == 'https'
            )
        except _SocketFail as error:
            self._socket_fail('unable to connect to proxy; {}', error)
        proxy_request = proxy.build_request(
            self.websocket.host, self.websocket.port,
            proxy_username=_proxy_url.username,
            proxy_password=_proxy_url.password
        )
        sock.sendall(proxy_request)
        proxy_parser = proxy.ProxyParser()
        response = None
        while response is None:
            data = sock.recv(1024)
            for response in proxy_parser.feed(data):
                break
        return (
            self._wrap_socket(sock, self.websocket.host)
            if self.websocket.is_secure else
            sock
        )

    def _connect(self):
        """Create socket and connect."""
        proxy = self.websocket.proxies.get(
            'https' if self.websocket.is_secure else 'http'
        )
        if proxy:
            sock = self._connect_proxy(proxy)
            proxy_url = proxy
        else:
            sock = self._connect_sock(
                self.websocket.host,
                self.websocket.port,
                ssl=self.websocket.is_secure
            )
            proxy_url = None
        # The timeout makes the socket non-blocking
        # We want to the socket to block after the connection
        sock.settimeout(None)
        return sock, proxy_url

    def _wrap_socket(self, sock, host):
        """Wrap the socket with an SSL proxy."""
        # sniff SNI support (added Python 2.7.9)
        if HAS_SNI:
            _protocol = getattr(
                ssl,
                'PROTOCOL_TLS',  # Supported since 2.7.13
                ssl.PROTOCOL_SSLv23   # Supported since 2.7.9
            )
            ssl_context = ssl.SSLContext(_protocol)
            ssl_sock = ssl_context.wrap_socket(
                sock, server_hostname=host
            )
        else:
            # Fallback for no SNI
            ssl_sock = ssl.wrap_socket(sock)
        log.debug('wrapped socket %r', ssl_sock)
        return ssl_sock

    def _close_socket(self):
        """Close the socket safely."""
        # Is a no-op if the socket is already closed.
        if self._sock is None:
            return
        try:
            # Get the write lock, so we can be certain data sending
            # in another thread is sent.
            with self._lock:
                self._sock.shutdown(socket.SHUT_RDWR)
                self._sock.close()
        except socket.error:
            # Socket is already closed, just a no-op
            pass
        except Exception as error:
            # Paranoia
            log.warning('error closing socket; %s', error)
        finally:
            self._sock = None

    def _send_request(self):
        """Send the request over the wire."""
        self.write(self.websocket.build_request())

    def _check_poll(self, poll, session_time):
        """Check if it is time for a poll."""
        _time = session_time
        if self._poll_start is None or _time - self._poll_start >= poll:
            self._poll_start = _time
            return True
        else:
            return False

    def _check_auto_ping(self, ping_rate, session_time):
        """Check if a ping is required."""
        if ping_rate and session_time > self._next_ping:
            # Calculate next ping time that is in the future.
            self._next_ping = (
                math.ceil(session_time / ping_rate) * ping_rate
            )
            try:
                self.websocket.send_ping()
            except errors.WebSocketError:
                pass  # If the websocket has gone away

    def _check_ping_timeout(self, ping_timeout, session_time):
        """Check if the server is not responding to pings."""
        if ping_timeout:
            time_since_last_pong = session_time - self._last_pong
            if time_since_last_pong > ping_timeout:
                log.debug('ping_timeout time exceeded')
                return True
        return False

    def _check_close_timeout(self, close_timeout, session_time):
        """Check if the close timeout was tripped."""
        if close_timeout:
            sent_close_time = self.websocket.sent_close_time
            if sent_close_time is None:
                return
            if session_time >= sent_close_time + close_timeout:
                raise _ForceDisconnect(
                    "server didn't respond to close packet "
                    "within {}s".format(close_timeout)
                )

    def _recv(self, count):
        """Receive and return pending data from the socket."""
        if self._sock is None:
            return bytearray(b'')
        try:
            _recv_count = self._sock.recv_into(self._buffer, count)
            return memoryview(self._buffer)[:_recv_count]
        except socket.error as error:
            log.debug('error in _recv', exc_info=True)
            self._socket_fail('recv fail; {}', error)

    def _regular(self, poll, ping_rate, ping_timeout, close_timeout):
        """Run regularly to do polling / pings."""
        # Check for regularly running actions.
        if self._check_poll(poll, self.session_time):
            yield events.Poll()
        self._check_auto_ping(ping_rate, self.session_time)
        if self._check_ping_timeout(ping_timeout, self.session_time):
            yield events.Unresponsive()
            raise _ForceDisconnect(
                'exceeded {:.0f}s ping timeout'.format(ping_timeout)
            )
        self._check_close_timeout(close_timeout, self.session_time)

    def _send_pong(self, event):
        """Send a pong message in response to ping event."""
        try:
            self.websocket.send_pong(event.data)
        except errors.WebSocketError:
            # In case the websocket has gone away
            pass

    def _on_pong(self, event):
        """Record last pong time."""
        self._last_pong = self.session_time

    def _on_ready(self):
        """Called when a ready event is received."""
        self._last_pong = 0.0
        self._next_ping = 0.0
        self._start_time = time.time()

    def _on_event(self, event, auto_pong=True):
        """Handle logic in response to an event."""
        if event.name == 'ready':
            self._on_ready()
            self._ready = True
        elif event.name == 'ping':
            if auto_pong:
                self._send_pong(event)
        elif event.name == 'pong':
            self._on_pong(event)

    def run(self,
            poll=5,
            ping_rate=30,
            ping_timeout=None,
            auto_pong=True,
            close_timeout=None):
        """Run the websocket."""
        websocket = self.websocket
        url = websocket.url
        # Connecting event
        yield events.Connecting(url)

        # Create socket and connect to remote server
        try:
            sock, proxy = self._connect()
            self._sock = sock
        except _SocketFail as error:
            yield events.ConnectFail('{}'.format(error))
            return
        except Exception as error:
            log.error('error connecting to %s; %s', url, error)
            yield events.ConnectFail('{}'.format(error))
            return

        # We now have a socket.
        # Send the request.
        try:
            self._send_request()
        except errors.WebSocketError as error:
            self._close_socket()
            yield events.ConnectFail('request failed; {}'.format(error))
            return

        # Connected to the server, but not yet upgraded to websockets
        yield events.Connected(url, proxy=proxy)

        selector = self._selector_cls(sock)
        log.debug('%r created', selector)

        def _regular():
            """Run regular events if websocket is ready."""
            if self._ready:
                return self._regular(
                    poll, ping_rate, ping_timeout, close_timeout
                )
            return ()

        try:
            while not websocket.is_closed:
                readable, max_bytes = selector.wait(self.BUFFER_SIZE, poll)
                for event in _regular():
                    yield event
                if readable:
                    data = self._recv(max_bytes)
                    if data:
                        for event in self.websocket.feed(data):
                            self._on_event(event, auto_pong)
                            yield event
                            for event in _regular():
                                yield event
                    else:
                        if websocket.is_active:
                            self._socket_fail('connection lost')
                        break
        except _ForceDisconnect as error:
            self._close_socket()
            yield events.Disconnected('disconnected; {}'.format(error))
        except _SocketFail as error:
            # Session methods will translate socket errors to this
            # exception. The result is we are disconnected.
            self._close_socket()
            yield events.Disconnected('socket fail; {}'.format(error))
        except Exception as error:  # pragma: no cover
            # It pays to be paranoid.
            log.exception('error in websocket loop')
            self._close_socket()
            yield events.Disconnected('error; {}'.format(error))
        else:
            # The websocket instance terminated the loop, which means
            # it was a graceful exit.
            self._close_socket()
            yield events.Disconnected(graceful=True)
        finally:
            selector.close()



from __future__ import unicode_literals

import select


class SelectorBase(object):
    """Abstraction for a kernel object that waits for socket data."""

    def __init__(self, socket):
        """Construct with an open socket."""
        self._socket = socket

    def wait(self, max_bytes, timeout=0.0):
        """Block until socket is readable or a timeout occurs. Return
        a tuple of <readable>,  <max bytes>.

        """
        if hasattr(self._socket, 'pending') and self._socket.pending():
            return True, self._socket.pending()
        readable = self.wait_readable(timeout=timeout)
        return readable, max_bytes

    def wait_readable(self, timeout=0.0):
        """Block until socket is readable or a timeout occurs, return
        `True` if the socket is readable, or `False` if the timeout
        occurred.

        """

    def close(self):
        """Close the selector (not the socket)."""


class SelectSelector(SelectorBase):  # pragma: no cover
    """Select Selector for use on Windows."""

    def __repr__(self):
        return '<SelectSelector>'

    def wait_readable(self, timeout=0.0):
        rlist, _wlist, _xlist = (
            select.select([self._socket.fileno()], [], [], timeout)
        )
        return bool(rlist)


class KQueueSelector(SelectorBase):  # pragma: no cover
    """KQueue selector for MacOS & BSD"""
    def __init__(self, socket):
        super(KQueueSelector, self).__init__(socket)
        self._queue = select.kqueue()
        self._events = [
            select.kevent(
                self._socket.fileno(),
                filter=select.KQ_FILTER_READ
            )
        ]

    def __repr__(self):
        return '<KQueueSelector>'

    def wait_readable(self, timeout=0.0):
        events = self._queue.control(
            self._events, 1, timeout
        )
        return bool(events)

    def close(self):
        self._queue.close()


class PollSelector(SelectorBase):
    """Poll selector for *nix"""
    def __init__(self, socket):
        super(PollSelector, self).__init__(socket)
        self._poll = select.poll()
        events = (
            select.POLLIN |
            select.POLLPRI |
            select.POLLERR |
            select.POLLHUP
        )
        self._poll.register(socket.fileno(), events)

    def __repr__(self):
        return '<PollSelector>'

    def wait_readable(self, timeout=0.0):
        events = self._poll.poll(timeout * 1000.0)
        return bool(events)


# Pick the appropriate selector for the given platform
if hasattr(select, 'kqueue'):
    PlatformSelector = KQueueSelector
elif hasattr(select, 'poll'):
    PlatformSelector = PollSelector
else:
    PlatformSelector = SelectSelector



"""
Parse a stream of Websocket frames, and optional HTTP headers.

"""

from __future__ import unicode_literals

import logging
import struct

from . import errors
from .frame import CompressedFrame, Frame
from .parser import Parser
from .utf8validator import Utf8Validator


log = logging.getLogger("lomond")


class FrameParser(Parser):
    """Parses a stream of data in to HTTP headers + WS frames."""

    def __init__(self, parse_headers=True, validate=True):
        self.parse_headers = parse_headers
        self.validate = validate
        self._is_text = False
        self._utf8_validator = Utf8Validator()
        self._frame_class = Frame
        self._compression = False
        super(FrameParser, self).__init__()

    def __repr__(self):
        return '{}(parser_headers={!r}, validate={!r})'.format(
            self.__class__.__name__,
            self.parse_headers,
            self.validate
        )

    # A bug in Python2.7.3 requires casting data to bytes
    @classmethod
    def unpack16(cls, data, _unpack16 = struct.Struct(b"!H").unpack):
        """Unpack 16 bits in to an integer."""
        return _unpack16(bytes(data))[0]

    @classmethod
    def unpack64(cls, data, _unpack64 = struct.Struct(b"!Q").unpack):
        """Unpack 64 bits in to an integer."""
        return _unpack64(bytes(data))[0]

    def enable_compression(self):
        """Enable compressed packets."""
        self._compression = True
        self._frame_class = CompressedFrame

    def read_text(self, length):
        """Read encoded text."""
        if self._compression:
            return self.read(length)
        else:
            return self.read_utf8(length, self._utf8_validator)

    def parse(self):
        # Get any WS frames
        if self.parse_headers:
            header_data = yield self.read_until(
                b"\r\n\r\n", max_bytes=16 * 1024
            )
            yield header_data

        while True:
            byte1, byte2 = yield self.read(2)

            fin = byte1 >> 7
            rsv1 = (byte1 >> 6) & 1
            rsv2 = (byte1 >> 5) & 1
            rsv3 = (byte1 >> 4) & 1
            opcode = byte1 & 0b00001111
            mask_bit = byte2 >> 7
            payload_length = byte2 & 0b01111111

            if payload_length == 126:
                payload_length = self.unpack16((yield self.read(2)))
            elif payload_length == 127:
                payload_length = self.unpack64((yield self.read(8)))
            if payload_length > 0x7fffffffffffffff:
                raise errors.PayloadTooLarge("payload is too large")

            if mask_bit:
                masking_key = yield self.read(4)
            else:
                masking_key = None

            frame = self._frame_class(
                opcode,
                fin=fin,
                rsv1=rsv1,
                rsv2=rsv2,
                rsv3=rsv3,
                mask=bool(mask_bit),
                masking_key=masking_key,
            )
            if self.validate:
                frame.validate()

            if frame.is_text:
                self._is_text = True

            if payload_length:
                _is_text_continuation = (
                    frame.is_continuation and self._is_text
                )
                if frame.is_text or _is_text_continuation:
                    frame.payload = yield self.read_text(payload_length)
                else:
                    frame.payload = yield self.read(payload_length)

            self.on_frame(frame)
            yield frame

    def on_frame(self, frame):
        """Called with new frames."""
        if (
            not self._compression
            and frame.fin
            and (frame.is_text or frame.is_continuation)
        ):
            self._utf8_validator.reset()
        if frame.fin:
            self._is_text = False


class ClientFrameParser(FrameParser):
    """Parse frames at client end."""

    def on_frame(self, frame):
        """Prohibit masked frames from server."""
        if frame.mask:
            log.warning(
                '%r must not have mask bit set',
                frame
            )
            raise errors.ProtocolError(
                'server sent masked frame'
            )
        super(ClientFrameParser, self).on_frame(frame)
